"""Helper methods for tests of custom Gym environments.

This is used in our test suite in `tests/test_envs.py`. It is also used in sister
projects such as `imitation`, and may be useful in other codebases.
"""

import re
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

import gym
import numpy as np

Step = Tuple[Any, Optional[float], bool, Mapping[str, Any]]
Rollout = Sequence[Step]
"""A sequence of 4-tuples (obs, rew, done, info) as returned by `get_rollout`."""


def make_env_fixture(
    skip_fn: Callable[[str], None],
) -> Callable[[str], Iterator[gym.Env]]:
    """Creates a fixture function, calling `skip_fn` when dependencies are missing.

    For example, in `pytest`, one would use::

        env = pytest.fixture(make_env_fixture(skip_fn=pytest.skip))

    Then any method with an `env` parameter will receive the created environment, with
    the `env_name` parameter automatically passed to the fixture.

    In `unittest`, one would use::

        def skip_fn(msg):
            raise unittest.SkipTest(msg)

        make_env = contextlib.contextmanager(make_env_fixture(skip_fn=skip_fn))

    And then call `with make_env(env_name) as env:` to create environments.

    Args:
        skip_fn: the function called when a dependency is missing to skip the test.

    Returns:
        A method to create Gym environments given their name.
    """

    def f(env_name: str) -> Iterator[gym.Env]:
        """Create environment `env_name`.

        Args:
            env_name: The name of the environment in the Gym registry.

        Yields:
            The created environment.

        Raises:
            gym.error.DependencyNotInstalled: if a dependency is missing
                other than MuJoCo (for MuJoCo, the test is instead skipped).
        """
        env = None
        try:
            env = gym.make(env_name)
            yield env
        except gym.error.DependencyNotInstalled as e:  # pragma: no cover
            if e.args[0].find("mujoco_py") != -1:
                skip_fn("Requires `mujoco_py`, which isn't installed.")
            else:
                raise
        finally:
            if env is not None:
                env.close()

    return f


def matches_list(env_name: str, patterns: Iterable[str]) -> bool:
    """Returns True if `env_name` matches any of the patterns in `patterns`."""
    return any(re.match(env_pattern, env_name) for env_pattern in patterns)


def get_rollout(env: gym.Env, actions: Iterable[Any]) -> Rollout:
    """Performs sequence of actions `actions` in `env`.

    Args:
      env: the environment to rollout in.
      actions: the actions to perform.

    Returns:
      A sequence of 4-tuples (obs, rew, done, info).
    """
    ret: List[Step] = [(env.reset(), None, False, {})]
    for act in actions:
        ret.append(env.step(act))
    return ret


def assert_equal_rollout(rollout_a: Rollout, rollout_b: Rollout) -> None:
    """Checks rollouts for equality.

    Raises:
        AssertionError if they are not equal.
    """
    for step_a, step_b in zip(rollout_a, rollout_b):
        ob_a, rew_a, done_a, info_a = step_a
        ob_b, rew_b, done_b, info_b = step_b
        np.testing.assert_equal(ob_a, ob_b)
        assert rew_a == rew_b
        assert done_a == done_b
        np.testing.assert_equal(info_a, info_b)


def has_same_observations(rollout_a: Rollout, rollout_b: Rollout) -> bool:
    """True if `rollout_a` and `rollout_b` have the same observations."""
    obs_list_a = [step[0] for step in rollout_a]
    obs_list_b = [step[0] for step in rollout_b]
    if len(obs_list_a) != len(obs_list_b):  # pragma: no cover
        return False
    for obs_a, obs_b in zip(obs_list_a, obs_list_b):
        if isinstance(obs_a, Mapping):  # pragma: no cover
            if obs_a.keys() != obs_b.keys():
                return False
            obs_a = list(obs_a.values())
            obs_b = list(obs_b.values())
        else:
            obs_a, obs_b = [obs_a], [obs_b]
        if any([np.any(x != y) for x, y in zip(obs_a, obs_b)]):
            return False
    return True


def test_seed(env: gym.Env, env_name: str, deterministic_envs: Iterable[str]) -> None:
    """Tests environment seeding.

    If non-deterministic, different seeds should produce different transitions.
    If deterministic, should be invariant to seed.

    Raises:
        AssertionError if test fails.
    """
    env.action_space.seed(0)
    actions = [env.action_space.sample() for _ in range(10)]

    # With the same seed, should always get the same result
    seeds = env.seed(42)
    assert isinstance(seeds, list)
    assert len(seeds) > 0
    rollout_a = get_rollout(env, actions)

    env.seed(42)
    rollout_b = get_rollout(env, actions)

    assert_equal_rollout(rollout_a, rollout_b)

    # For most non-deterministic environments, if we try enough seeds we should
    # eventually get a different result. For deterministic environments, all
    # seeds should produce the same starting state.
    def different_seeds_same_rollout(seed1, seed2):
        new_actions = [env.action_space.sample() for _ in range(10)]
        env.seed(seed1)
        new_rollout_1 = get_rollout(env, new_actions)
        env.seed(seed2)
        new_rollout_2 = get_rollout(env, new_actions)
        return has_same_observations(new_rollout_1, new_rollout_2)

    is_deterministic = matches_list(env_name, deterministic_envs)
    same_obs = all(different_seeds_same_rollout(seed, seed + 1) for seed in range(100))
    assert same_obs == is_deterministic


def _check_obs(obs: np.ndarray, obs_space: gym.Space) -> None:
    """Check obs is consistent with obs_space."""
    if obs_space.shape:
        assert obs.shape == obs_space.shape
        assert obs.dtype == obs_space.dtype
    assert obs in obs_space


def _sample_and_check(env: gym.Env, obs_space: gym.Space) -> bool:
    """Sample from env and check return value is of valid type."""
    act = env.action_space.sample()
    obs, rew, done, info = env.step(act)
    _check_obs(obs, obs_space)
    assert isinstance(rew, float)
    assert isinstance(done, bool)
    assert isinstance(info, dict)
    return done


def _is_mujoco_env(env: gym.Env) -> bool:
    return hasattr(env, "sim") and hasattr(env, "model")


def test_rollout_schema(
    env: gym.Env,
    steps_after_done: int = 10,
    max_steps: int = 10000,
) -> None:
    """Check custom environments have correct types on `step` and `reset`.

    Args:
        env: The environment to test.
        steps_after_done: The number of steps to take after `done` is True, the nominal
            episode termination. This is an abuse of the Gym API, but we would like the
            environments to handle this case gracefully.
        max_steps: Test fails if we do not get `done` after this many timesteps.

    Raises:
        AssertionError if test fails.
    """
    obs_space = env.observation_space
    obs = env.reset()
    _check_obs(obs, obs_space)

    for _ in range(max_steps):
        done = _sample_and_check(env, obs_space)
        if done:
            break

    assert done is True, "did not get to end of episode"

    for _ in range(steps_after_done):
        _sample_and_check(env, obs_space)


def test_premature_step(env: gym.Env, skip_fn, raises_fn) -> None:
    """Test that you must call reset() before calling step().

    Example usage in pytest:
        test_premature_step(env, skip_fn=pytest.skip, raises_fn=pytest.raises)

    Args:
        env: The environment to test.
        skip_fn: called when the environment is incompatible with the test.
        raises_fn: Context manager to check exception is thrown.

    Raises:
        AssertionError if test fails.
    """
    if _is_mujoco_env(env):  # pragma: no cover
        # We can't use isinstance since importing mujoco_py will fail on
        # machines without MuJoCo installed
        skip_fn("MuJoCo environments cannot perform this check.")

    act = env.action_space.sample()
    with raises_fn(Exception):  # need to call env.reset() first
        env.step(act)


def test_render(env: gym.Env, raises_fn) -> None:
    """Test that render() supports the modes declared.

    Example usage in pytest:
        test_render(env, raises_fn=pytest.raises)

    Args:
        env: The environment to test.
        raises_fn: Context manager to check NotImplementedError is thrown when
            environment metadata indicates modes are supported.

    Raises:
        AssertionError: if test fails. This occurs if:
            (a) `env.render(mode=mode)` fails for any mode declared supported
            in `env.metadata["render.modes"]`; (b) env.render() *succeeds* when
            `env.metadata["render.modes"]` is empty; (c) `env.render(mode="rgb_array")`
            returns different values at the same time step.
    """
    env.reset()  # make sure environment is in consistent state

    render_modes = env.metadata["render.modes"]
    if not render_modes:
        # No modes supported -- render() should fail.
        with raises_fn(NotImplementedError):
            env.render()
    else:
        for mode in render_modes:
            env.render(mode=mode)

        # WARNING(adam): there seems to be a memory leak with Gym 0.17.3
        # & MuJoCoPy 1.50.1.68. `MujocoEnv.close()` does not call `finish()`
        # on the viewer (commented out) so the resources are not released.
        # For now this is OK, but may bite if we end up testing a lot of
        # MuJoCo environments.
        is_mujoco = _is_mujoco_env(env)
        if "rgb_array" in render_modes and not is_mujoco:
            # Render should not change without calling `step()`.
            # MuJoCo rendering fails this check, ignore -- not much we can do.
            resa = env.render(mode="rgb_array")
            resb = env.render(mode="rgb_array")
            assert np.allclose(resa, resb)


class CountingEnv(gym.Env):
    """At timestep `t` of each episode, has `t == obs == reward / 10`.

    Episodes finish after `episode_length` calls to `step()`, or equivalently
    `episode_length` actions. For example, if we have `episode_length=5`,
    then an episode has the following observations and rewards:

    ```
    obs = [0, 1, 2, 3, 4, 5]
    rews = [10, 20, 30, 40, 50]
    ```
    """

    def __init__(self, episode_length: int = 5):
        """Initialize a CountingEnv.

        Params:
            episode_length: The number of actions before each episode ends.
        """
        assert episode_length >= 1
        self.observation_space = gym.spaces.Box(low=0, high=np.inf, shape=())
        self.action_space = gym.spaces.Box(low=0, high=np.inf, shape=())
        self.episode_length = episode_length
        self.timestep = None

    def reset(self):
        """Reset method for CountingEnv."""
        t, self.timestep = 0, 1
        return np.array(t, dtype=self.observation_space.dtype)

    def step(self, action):
        """Step method for CountingEnv."""
        if self.timestep is None:  # pragma: no cover
            raise RuntimeError("Need to reset before first step().")
        if np.array(action) not in self.action_space:  # pragma: no cover
            raise ValueError(f"Invalid action {action}")
        if self.timestep > self.episode_length:  # pragma: no cover
            raise ValueError("Should reset env. Episode is over.")

        t, self.timestep = self.timestep, self.timestep + 1
        obs = np.array(t, dtype=self.observation_space.dtype)
        rew = t * 10.0
        done = t == self.episode_length
        return obs, rew, done, {}



"""Environment with shift in initial state distribution."""

import functools
import itertools

import numpy as np

from seals import base_envs


class InitShiftEnv(base_envs.TabularModelMDP):
    """Tests for robustness to initial state shift.

    Many LfH algorithms learn from expert demonstrations. This can be
    problematic when the environment the demonstrations were gathered in
    differs even slightly from the learner's environment.

    This task illustrates this problem.  We have a depth-2 full binary tree
    where the agent moves left or right until reaching a leaf. The expert
    starts at the root s_0, whereas the learner starts at the left branch s_1
    and so can only reach leaves s_3 and s_4. Reward is only given at the
    leaves.

    The expert always move to the highest reward leaf s_6, so any algorithm
    that relies on demonstrations will not know whether it is better to go to
    s_3 or s_4. By contrast, feedback such as preference comparison can
    disambiguate this case.
    """

    def __init__(self, initial_state: int):
        """Constructs environment.

        Args:
            initial_state: fixed initial state.

        Raises:
            ValueError: `initial_state` not in [0,6].
        """
        nS = 7
        nA = 2

        if not 0 <= initial_state < nS:
            raise ValueError(f"Initial state {initial_state} must lie in [0,{nS})")

        self._initial_state = initial_state

        non_leaves = np.arange(3)
        leaves = np.arange(3, 7)

        transition_matrix = np.zeros((nS, nA, nS))

        for state, action in itertools.product(non_leaves, range(nA)):
            next_state = 2 * state + 1 + action
            transition_matrix[state, action, next_state] = 1.0

        transition_matrix[leaves, :, leaves] = 1.0

        reward_matrix = np.zeros((nS,))
        reward_matrix[leaves] = [1, -1, -1, 2]

        super().__init__(
            transition_matrix=transition_matrix,
            reward_matrix=reward_matrix,
        )

    def initial_state(self) -> int:
        """Returns initial state defined in constructor."""
        return self._initial_state


InitShiftTrainEnv = functools.partial(InitShiftEnv, initial_state=0)
InitShiftTestEnv = functools.partial(InitShiftEnv, initial_state=1)



"""Adaptation of classic Gym environments for specification learning algorithms."""

import warnings

from gym import spaces
import gym.envs.classic_control
import numpy as np

from seals import util


class FixedHorizonCartPole(gym.envs.classic_control.CartPoleEnv):
    """Fixed-length variant of CartPole-v1.

    Reward is 1.0 whenever the CartPole is an "ok" state (i.e. the pole is upright
    and the cart is on the screen). Otherwise reward is 0.0.

    Done is always False. (Though note that by default, this environment is wrapped
    in `TimeLimit` with max steps 500.)
    """

    def __init__(self):
        """Builds FixedHorizonCartPole, modifying observation_space from Gym parent."""
        super().__init__()

        high = [
            np.finfo(np.float32).max,  # x axis
            np.finfo(np.float32).max,  # x velocity
            np.pi,  # theta in radians
            np.finfo(np.float32).max,  # theta velocity
        ]
        high = np.array(high)
        self.observation_space = spaces.Box(-high, high, dtype=np.float32)

    def reset(self):
        """Reset for FixedHorizonCartPole."""
        return super().reset().astype(np.float32)

    def step(self, action):
        """Step function for FixedHorizonCartPole."""
        with warnings.catch_warnings():
            # Filter out CartPoleEnv warning for calling step() beyond done=True.
            warnings.filterwarnings("ignore", ".*You are calling.*")
            super().step(action)

        self.state = list(self.state)
        x, _, theta, _ = self.state

        # Normalize theta to [-pi, pi] range.
        theta = (theta + np.pi) % (2 * np.pi) - np.pi
        self.state[2] = theta

        state_ok = bool(
            abs(x) < self.x_threshold and abs(theta) < self.theta_threshold_radians,
        )

        rew = 1.0 if state_ok else 0.0
        return np.array(self.state, dtype=np.float32), rew, False, {}


def mountain_car():
    """Fixed-length variant of MountainCar-v0.

    In the event of early episode completion (i.e., the car reaches the
    goal), we enter an absorbing state that repeats the final observation
    and returns reward 0.

    Done is always returned on timestep 200 only.
    """
    env = util.make_env_no_wrappers("MountainCar-v0")
    env = util.ObsCastWrapper(env, dtype=np.float32)
    env = util.AbsorbAfterDoneWrapper(env)
    return env



"""Adaptation of MuJoCo environments for specification learning algorithms."""

import functools

from gym.envs.mujoco import (
    ant_v3,
    half_cheetah_v3,
    hopper_v3,
    humanoid_v3,
    swimmer_v3,
    walker2d_v3,
)


def _include_position_in_observation(cls):
    cls.__init__ = functools.partialmethod(
        cls.__init__,
        exclude_current_positions_from_observation=False,
    )
    return cls


def _no_early_termination(cls):
    cls.__init__ = functools.partialmethod(cls.__init__, terminate_when_unhealthy=False)
    return cls


@_include_position_in_observation
@_no_early_termination
class AntEnv(ant_v3.AntEnv):
    """Ant with position observation and no early termination."""


@_include_position_in_observation
class HalfCheetahEnv(half_cheetah_v3.HalfCheetahEnv):
    """HalfCheetah with position observation and no early termination."""


@_include_position_in_observation
@_no_early_termination
class HopperEnv(hopper_v3.HopperEnv):
    """Hopper with position observation and no early termination."""


@_include_position_in_observation
@_no_early_termination
class HumanoidEnv(humanoid_v3.HumanoidEnv):
    """Humanoid with position observation and no early termination."""


@_include_position_in_observation
class SwimmerEnv(swimmer_v3.SwimmerEnv):
    """Swimmer with position observation. Naturally does not terminate early."""


@_include_position_in_observation
@_no_early_termination
class Walker2dEnv(walker2d_v3.Walker2dEnv):
    """Walker2d with position observation and no early termination."""



"""Environment checking for correctness under early termination."""

import functools

import numpy as np

from seals import base_envs


class EarlyTerminationEnv(base_envs.TabularModelMDP):
    """Three-state MDP with early termination state.

    Many implementations of imitation learning algorithms incorrectly assign a
    value of zero to terminal states [1]. Depending on the sign of the learned
    reward function in non-terminal states, this can either bias the agent to
    end episodes early or prolong them as long as possible. This confounds
    evaluation as performance is spuriously high in tasks where the termination
    bias aligns with the task objective. These tasks attempt to detect this
    type of bias, and they are adapted from [1].

    The environment is a 3-state MDP, in which the agent can either alternate
    between two initial states until reaching the time horizon, or they can
    move to a terminal state causing the episode to terminate early.

    [1] Kostrikov, Ilya, et al. "Discriminator-actor-critic: Addressing sample
    inefficiency and reward bias in adversarial imitation learning." arXiv
    preprint arXiv:1809.02925 (2018).
    """

    def __init__(self, is_reward_positive: bool = True):
        """Construct environment.

        Args:
            is_reward_positive: whether rewards are positive or negative.
        """
        nS = 3
        nA = 2

        transition_matrix = np.zeros((nS, nA, nS))

        transition_matrix[0, :, 1] = 1.0

        transition_matrix[1, 0, 0] = 1.0
        transition_matrix[1, 1, 2] = 1.0

        transition_matrix[2, :, 2] = 1.0

        reward_sign = 2 * is_reward_positive - 1
        reward_matrix = reward_sign * np.ones((nS,), dtype=float)

        super().__init__(
            transition_matrix=transition_matrix,
            reward_matrix=reward_matrix,
        )

    def terminal(self, state: int, n_actions_taken: int) -> bool:
        """Returns True if (and only if) in state 2."""
        return bool(state == 2)


EarlyTermPosEnv = functools.partial(EarlyTerminationEnv, is_reward_positive=True)
EarlyTermNegEnv = functools.partial(EarlyTerminationEnv, is_reward_positive=False)



"""Environment testing for generalization in continuous spaces."""

from gym import spaces
import numpy as np

from seals import base_envs


class ParabolaEnv(base_envs.ResettableMDP):
    """Environment to mimic parabola curves.

    This environment tests algorithms' ability to learn in continuous
    action spaces, a challenge for Q-learning methods in particular.
    The goal is to mimic the path of a parabola p(x) = A*x**2 + B*x +
    C, where A, B and C are constants sampled uniformly from [-1, 1]
    at the start of the episode.
    """

    def __init__(self, x_step: float = 0.05, bounds: float = 5):
        """Construct environment.

        Args:
            x_step: x position difference between timesteps.
            bounds: limits coordinates, useful for keeping rewards in
                a small bounded range.
        """
        self._x_step = x_step
        self._bounds = bounds

        state_high = np.array([bounds, bounds, 1.0, 1.0, 1.0])
        state_low = (-1) * state_high

        super().__init__(
            state_space=spaces.Box(low=state_low, high=state_high),
            action_space=spaces.Box(low=-np.inf, high=np.inf, shape=()),
        )

    def terminal(self, state: int, n_actions_taken: int) -> bool:
        """Always returns False."""
        return False

    def initial_state(self) -> np.ndarray:
        """Get state by sampling a random parabola."""
        a, b, c = -1 + 2 * self.rand_state.rand(3)
        x, y = 0, c
        return np.array([x, y, a, b, c], dtype=self.state_space.dtype)

    def reward(self, state: np.ndarray, action: int, new_state: np.ndarray) -> float:
        """Negative squared vertical distance from parabola."""
        x, y, a, b, c = state
        target_y = a * x ** 2 + b * x + c
        return (-1) * (y - target_y) ** 2

    def transition(self, state: np.ndarray, action: int) -> np.ndarray:
        """Update x according to x_step and y according to action."""
        x, y, a, b, c = state
        next_x = np.clip(x + self._x_step, -self._bounds, self._bounds)
        next_y = np.clip(y + action, -self._bounds, self._bounds)
        return np.array([next_x, next_y, a, b, c], dtype=self.state_space.dtype)



"""Simple diagnostic environments."""

import gym

gym.register(
    id="seals/Branching-v0",
    entry_point="seals.diagnostics.branching:BranchingEnv",
    max_episode_steps=11,
)

gym.register(
    id="seals/EarlyTermNeg-v0",
    entry_point="seals.diagnostics.early_term:EarlyTermNegEnv",
    max_episode_steps=10,
)

gym.register(
    id="seals/EarlyTermPos-v0",
    entry_point="seals.diagnostics.early_term:EarlyTermPosEnv",
    max_episode_steps=10,
)

gym.register(
    id="seals/InitShiftTrain-v0",
    entry_point="seals.diagnostics.init_shift:InitShiftTrainEnv",
    max_episode_steps=3,
)

gym.register(
    id="seals/InitShiftTest-v0",
    entry_point="seals.diagnostics.init_shift:InitShiftTestEnv",
    max_episode_steps=3,
)

gym.register(
    id="seals/LargestSum-v0",
    entry_point="seals.diagnostics.largest_sum:LargestSumEnv",
    max_episode_steps=1,
)

gym.register(
    id="seals/NoisyObs-v0",
    entry_point="seals.diagnostics.noisy_obs:NoisyObsEnv",
    max_episode_steps=15,
)

gym.register(
    id="seals/Parabola-v0",
    entry_point="seals.diagnostics.parabola:ParabolaEnv",
    max_episode_steps=20,
)

gym.register(
    id="seals/ProcGoal-v0",
    entry_point="seals.diagnostics.proc_goal:ProcGoalEnv",
    max_episode_steps=20,
)

gym.register(
    id="seals/RiskyPath-v0",
    entry_point="seals.diagnostics.risky_path:RiskyPathEnv",
    max_episode_steps=5,
)

gym.register(
    id="seals/Sort-v0",
    entry_point="seals.diagnostics.sort:SortEnv",
    max_episode_steps=6,
)



"""Base environment classes."""

import abc
from typing import Generic, Optional, Sequence, Tuple, TypeVar

import gym
from gym import spaces
import numpy as np

from seals import util

State = TypeVar("State")
Observation = TypeVar("Observation")
Action = TypeVar("Action")


class ResettablePOMDP(gym.Env, abc.ABC, Generic[State, Observation, Action]):
    """ABC for POMDPs that are resettable.

    Specifically, these environments provide oracle access to sample from
    the initial state distribution and transition dynamics, and compute the
    reward and termination condition. Almost all simulated environments can
    meet these criteria.
    """

    def __init__(
        self,
        *,
        state_space: gym.Space,
        observation_space: gym.Space,
        action_space: gym.Space,
    ):
        """Build resettable (PO)MDP.

        Args:
            state_space: gym.Space containing possible states.
            observation_space: gym.Space containing possible observations.
            action_space: gym.Space containing possible actions.
        """
        self._state_space = state_space
        self._observation_space = observation_space
        self._action_space = action_space

        self.cur_state: Optional[State] = None
        self._n_actions_taken: Optional[int] = None
        self.seed()

    @abc.abstractmethod
    def initial_state(self) -> State:
        """Samples from the initial state distribution."""

    @abc.abstractmethod
    def transition(self, state: State, action: Action) -> State:
        """Samples from transition distribution."""

    @abc.abstractmethod
    def reward(self, state: State, action: Action, new_state: State) -> float:
        """Computes reward for a given transition."""

    @abc.abstractmethod
    def terminal(self, state: State, step: int) -> bool:
        """Is the state terminal?"""

    @abc.abstractmethod
    def obs_from_state(self, state: State) -> Observation:
        """Sample observation for given state."""

    @property
    def state_space(self) -> gym.Space:
        """State space. Often same as observation_space, but differs in POMDPs."""
        return self._state_space

    @property
    def observation_space(self) -> gym.Space:
        """Observation space. Return type of reset() and component of step()."""
        return self._observation_space

    @property
    def action_space(self) -> gym.Space:
        """Action space. Parameter type of step()."""
        return self._action_space

    @property
    def n_actions_taken(self) -> int:
        """Number of steps taken so far."""
        assert self._n_actions_taken is not None
        return self._n_actions_taken

    def seed(self, seed=None) -> Sequence[int]:
        """Set random seed."""
        if seed is None:
            # Gym API wants list of seeds to be returned for some reason, so
            # generate a seed explicitly in this case
            seed = np.random.randint(0, 1 << 31)
        self.rand_state = np.random.RandomState(seed)
        return [seed]

    def reset(self) -> Observation:
        """Reset episode and return initial observation."""
        self.cur_state = self.initial_state()
        assert self.cur_state in self.state_space, f"unexpected state {self.cur_state}"
        self._n_actions_taken = 0
        return self.obs_from_state(self.cur_state)

    def step(self, action: Action) -> Tuple[Observation, float, bool, dict]:
        """Transition state using given action."""
        if self.cur_state is None or self._n_actions_taken is None:
            raise ValueError("Need to call reset() before first step()")
        if action not in self.action_space:
            raise ValueError(f"{action} not in {self.action_space}")

        old_state = self.cur_state
        self.cur_state = self.transition(self.cur_state, action)
        assert self.cur_state in self.state_space, f"unexpected state {self.cur_state}"
        obs = self.obs_from_state(self.cur_state)
        assert obs in self.observation_space, f"{obs} not in {self.observation_space}"
        rew = self.reward(old_state, action, self.cur_state)
        done = self.terminal(self.cur_state, self._n_actions_taken)
        self._n_actions_taken += 1

        infos = {"old_state": old_state, "new_state": self.cur_state}
        return obs, rew, done, infos


class ResettableMDP(ResettablePOMDP[State, State, Action], Generic[State, Action]):
    """ABC for MDPs that are resettable."""

    def __init__(
        self,
        *,
        state_space: gym.Space,
        action_space: gym.Space,
    ):
        """Build resettable MDP.

        Args:
            state_space: gym.Space containing possible states.
            action_space: gym.Space containing possible actions.
        """
        super().__init__(
            state_space=state_space,
            observation_space=state_space,
            action_space=action_space,
        )

    def obs_from_state(self, state: State) -> State:
        """Identity since observation == state in an MDP."""
        return state


class TabularModelMDP(ResettableMDP[int, int]):
    """Base class for tabular environments with known dynamics."""

    def __init__(
        self,
        *,
        transition_matrix: np.ndarray,
        reward_matrix: np.ndarray,
        horizon: float = np.inf,
        initial_state_dist: Optional[np.ndarray] = None,
    ):
        """Build tabular environment.

        Args:
            transition_matrix: 3-D array with transition probabilities for a
                given state-action pair, of shape `(n_states,n_actions,n_states)`.
            reward_matrix: 1-D, 2-D or 3-D array corresponding to rewards to a
                given `(state, action, next_state)` triple. A 2-D array assumes
                the `next_state` is not used in the reward, and a 1-D array
                assumes neither the `action` nor `next_state` are used.
                Of shape `(n_states,n_actions,n_states)[:n]` where `n`
                is the dimensionality of the array.
            horizon: Maximum number of timesteps, default `np.inf`.
            initial_state_dist: Distribution from which state is sampled at the
                start of the episode.  If `None`, it is assumed initial state
                is always 0. Shape `(n_states,)`.

        Raises:
            ValueError: `transition_matrix`, `reward_matrix` or
                `initial_state_dist` have shapes different to specified above.
        """
        n_states, n_actions, n_next_states = transition_matrix.shape
        if n_states != n_next_states:
            raise ValueError(
                "Malformed transition_matrix:\n"
                f"transition_matrix.shape: {transition_matrix.shape}\n"
                f"{n_states} != {n_next_states}",
            )

        if initial_state_dist is None:
            initial_state_dist = util.one_hot_encoding(0, n_states)
        if initial_state_dist.ndim != 1:
            raise ValueError(
                "initial_state_dist has multiple dimensions:\n"
                f"{initial_state_dist.ndim} != 1",
            )
        if initial_state_dist.shape[0] != n_states:
            raise ValueError(
                "transition_matrix and initial_state_dist are not compatible:\n"
                f"n_states = {n_states}\n"
                f"len(initial_state_dist) = {len(initial_state_dist)}",
            )

        if reward_matrix.shape != transition_matrix.shape[: len(reward_matrix.shape)]:
            raise ValueError(
                "transition_matrix and reward_matrix are not compatible:\n"
                f"transition_matrix.shape: {transition_matrix.shape}\n"
                f"reward_matrix.shape: {reward_matrix.shape}",
            )

        self.transition_matrix = transition_matrix
        self.reward_matrix = reward_matrix
        self._feature_matrix = None
        self.horizon = horizon
        self.initial_state_dist = initial_state_dist

        super().__init__(
            state_space=spaces.Discrete(n_states),
            action_space=spaces.Discrete(n_actions),
        )

    def initial_state(self) -> int:
        """Samples from the initial state distribution."""
        return util.sample_distribution(
            self.initial_state_dist,
            random=self.rand_state,
        )

    def transition(self, state: int, action: int) -> int:
        """Samples from transition distribution."""
        return util.sample_distribution(
            self.transition_matrix[state, action],
            random=self.rand_state,
        )

    def reward(self, state: int, action: int, new_state: int) -> float:
        """Computes reward for a given transition."""
        inputs = (state, action, new_state)[: len(self.reward_matrix.shape)]
        return self.reward_matrix[inputs]

    def terminal(self, state: int, n_actions_taken: int) -> bool:
        """Checks if state is terminal."""
        return n_actions_taken >= self.horizon

    @property
    def feature_matrix(self):
        """Matrix mapping states to feature vectors."""
        # Construct lazily to save memory in algorithms that don't need features.
        if self._feature_matrix is None:
            n_states = self.state_space.n
            self._feature_matrix = np.eye(n_states)
        return self._feature_matrix



"""Large gridworld with random agent and goal position."""

from gym import spaces
import numpy as np

from seals import base_envs, util


class ProcGoalEnv(base_envs.ResettableMDP):
    """Large gridworld with random agent and goal position.

    In this task, the agent starts at a random position in a large
    grid, and must navigate to a goal randomly placed in a
    neighborhood around the agent.  The observation is a 4-dimensional
    vector containing the (x,y) coordinates of the agent and the goal.
    The reward at each timestep is the negative Manhattan distance
    between the two positions.  With a large enough grid, generalizing
    is necessary to achieve good performance, since most initial
    states will be unseen.
    """

    def __init__(self, bounds: int = 100, distance: int = 10):
        """Constructs environment.

        Args:
            bounds: the absolute values of the coordinates of the initial agent
                position are bounded by `bounds`. Increasing the value might make
                generalization harder.
            distance: initial distance between agent and goal.
        """
        self._bounds = bounds
        self._distance = distance

        super().__init__(
            state_space=spaces.Box(low=-np.inf, high=np.inf, shape=(4,)),
            action_space=spaces.Discrete(5),
        )

    def terminal(self, state: np.ndarray, n_actions_taken: int) -> bool:
        """Always returns False."""
        return False

    def initial_state(self) -> np.ndarray:
        """Samples random agent position and random goal."""
        pos = self.rand_state.randint(low=-self._bounds, high=self._bounds, size=(2,))

        x_dist = self.rand_state.randint(self._distance)
        y_dist = self._distance - x_dist
        random_signs = 2 * self.rand_state.randint(2, size=2) - 1
        goal = pos + random_signs * (x_dist, y_dist)

        return np.concatenate([pos, goal]).astype(self.observation_space.dtype)

    def reward(self, state: np.ndarray, action: int, new_state: np.ndarray) -> float:
        """Negative L1 distance to goal."""
        return (-1) * np.sum(np.abs(state[2:] - state[:2]))

    def transition(self, state: np.ndarray, action: int) -> np.ndarray:
        """Returns next state according to grid."""
        pos, goal = state[:2], state[2:]
        next_pos = util.grid_transition_fn(pos, action)
        return np.concatenate([next_pos, goal])



"""Miscellaneous utilities."""

from typing import Optional, Tuple

import gym
import numpy as np


class AutoResetWrapper(gym.Wrapper):
    """Hides done=True and auto-resets at the end of each episode."""

    def step(self, action):
        """When done=True, returns done=False instead and automatically resets.

        When an automatic reset happens, the observation from reset is returned,
        and the overridden observation is stored in
        `info["terminal_observation"]`.
        """
        obs, rew, done, info = self.env.step(action)
        if done:
            info["terminal_observation"] = obs
            obs = self.env.reset()
        return obs, rew, False, info


class ObsCastWrapper(gym.Wrapper):
    """Cast observations to specified dtype.

    Some external environments return observations of a different type than the
    declared observation space. Where possible, this should be fixed upstream,
    but casting can be a viable workaround -- especially when the returned
    observations are higher resolution than the observation space.
    """

    def __init__(self, env: gym.Env, dtype: np.dtype):
        """Builds ObsCastWrapper.

        Args:
            env: the environment to wrap.
            dtype: the dtype to cast observations to.
        """
        super().__init__(env)
        self.dtype = dtype

    def reset(self):
        """Returns reset observation, cast to self.dtype."""
        return super().reset().astype(self.dtype)

    def step(self, action):
        """Returns (obs, rew, done, info) with obs cast to self.dtype."""
        obs, rew, done, info = super().step(action)
        return obs.astype(self.dtype), rew, done, info


class AbsorbAfterDoneWrapper(gym.Wrapper):
    """Transition into absorbing state instead of episode termination.

    When the environment being wrapped returns `done=True`, we return an absorbing
    observation. This wrapper always returns `done=False`.

    A convenient way to add absorbing states to environments like MountainCar.
    """

    def __init__(
        self,
        env: gym.Env,
        absorb_reward: float = 0.0,
        absorb_obs: Optional[np.ndarray] = None,
    ):
        """Initialize AbsorbAfterDoneWrapper.

        Args:
          env: The wrapped Env.
          absorb_reward: The reward returned at the absorb state.
          absorb_obs: The observation returned at the absorb state. If None, then
            repeat the final observation before absorb.
        """
        super().__init__(env)
        self.absorb_reward = absorb_reward
        self.absorb_obs_default = absorb_obs
        self.absorb_obs_this_episode = None
        self.at_absorb_state = None

    def reset(self, *args, **kwargs):
        """Reset the environment."""
        self.at_absorb_state = False
        self.absorb_obs_this_episode = None
        return self.env.reset(*args, **kwargs)

    def step(self, action):
        """Advance the environment by one step.

        This wrapped `step()` always returns done=False.

        After the first done is returned by the underlying Env, we enter an artificial
        absorb state.

        In this artificial absorb state, we stop calling
        `self.env.step(action)` (i.e. the `action` argument is entirely ignored) and
        we return fixed values for obs, rew, done, and info. The values of `obs` and
        `rew` depend on initialization arguments. `info` is always an empty dictionary.
        """
        if not self.at_absorb_state:
            inner_obs, inner_rew, done, inner_info = self.env.step(action)
            if done:
                # Initialize the artificial absorb state, which we will repeatedly use
                # starting on the next call to `step()`.
                self.at_absorb_state = True

                if self.absorb_obs_default is None:
                    self.absorb_obs_this_episode = inner_obs
                else:
                    self.absorb_obs_this_episode = self.absorb_obs_default
            obs, rew, info = inner_obs, inner_rew, inner_info
        else:
            assert self.absorb_obs_this_episode is not None
            assert self.absorb_reward is not None
            obs = self.absorb_obs_this_episode
            rew = self.absorb_reward
            info = {}

        return obs, rew, False, info


def make_env_no_wrappers(env_name: str, **kwargs) -> gym.Env:
    """Gym sometimes wraps envs in TimeLimit before returning from gym.make().

    This helper method builds directly from spec to avoid this wrapper.
    """
    return gym.envs.registry.env_specs[env_name].make(**kwargs)


def get_gym_max_episode_steps(env_name: str) -> Optional[int]:
    """Get the `max_episode_steps` attribute associated with a gym Spec."""
    return gym.envs.registry.env_specs[env_name].max_episode_steps


def sample_distribution(
    p: np.ndarray,
    random: np.random.RandomState,
) -> int:
    """Samples an integer with probabilities given by p."""
    return random.choice(np.arange(len(p)), p=p)


def one_hot_encoding(pos: int, size: int) -> np.ndarray:
    """Returns a 1-D hot encoding of a given position and size."""
    return np.eye(size)[pos]


def grid_transition_fn(
    state: np.ndarray,
    action: int,
    x_bounds: Tuple[float, float] = (-np.inf, np.inf),
    y_bounds: Tuple[float, float] = (-np.inf, np.inf),
):
    """Returns transition of a deterministic gridworld.

    Agent is bounded in the region limited by x_bounds and y_bounds,
    ends inclusive.

    (0, 0) is interpreted to be top-left corner.

    Actions:
    0: Right
    1: Down
    2: Left
    3: Up
    4: Stay put
    """
    dirs = [
        (1, 0),
        (0, 1),
        (-1, 0),
        (0, -1),
        (0, 0),
    ]

    x, y = state
    dx, dy = dirs[action]

    next_x = np.clip(x + dx, *x_bounds)
    next_y = np.clip(y + dy, *y_bounds)
    next_state = np.array([next_x, next_y], dtype=state.dtype)

    return next_state



"""Hard-exploration environment."""

import itertools

import numpy as np

from seals import base_envs


class BranchingEnv(base_envs.TabularModelMDP):
    """Long branching environment requiring exploration.

    The agent must traverse a specific path of length L to reach a
    final goal, with B choices at each step. Wrong actions lead to
    dead-ends with zero reward.
    """

    def __init__(self, branch_factor: int = 2, length: int = 10):
        """Construct environment.

        Args:
            branch_factor: number of actions at each state.
            length: path length from initial state to goal.
        """
        nS = 1 + branch_factor * length
        nA = branch_factor

        def get_next(state: int, action: int) -> int:
            can_move = state % branch_factor == 0 and state != nS - 1
            return state + (action + 1) * can_move

        transition_matrix = np.zeros((nS, nA, nS))
        for state, action in itertools.product(range(nS), range(nA)):
            transition_matrix[state, action, get_next(state, action)] = 1.0

        reward_matrix = np.zeros((nS,))
        reward_matrix[-1] = 1.0

        super().__init__(
            transition_matrix=transition_matrix,
            reward_matrix=reward_matrix,
        )



"""Environment testing scalability to high-dimensional tasks."""

from gym import spaces
import numpy as np

from seals import base_envs


class LargestSumEnv(base_envs.ResettableMDP):
    """High-dimensional linear classification problem.

    This environment evaluates how algorithms scale with increasing
    dimensionality.  It is a classification task with binary actions
    and uniformly sampled states s in [0, 1]**L.  The agent is
    rewarded for taking action 1 if the sum of the first half x[:L//2]
    is greater than the sum of the second half x[L//2:], and otherwise
    is rewarded for taking action 0.
    """

    def __init__(self, length: int = 50):
        """Build environment.

        Args:
            length: dimensionality of state space vector.
        """
        self._length = length
        state_space = spaces.Box(low=0.0, high=1.0, shape=(length,))
        super().__init__(
            state_space=state_space,
            action_space=spaces.Discrete(2),
        )

    def terminal(self, state: np.ndarray, n_actions_taken: int) -> bool:
        """Always returns True, since this task should have a 1-timestep horizon."""
        return True

    def initial_state(self) -> np.ndarray:
        """Returns vector sampled uniformly in [0, 1]**L."""
        init_state = self.rand_state.rand(self._length)
        return init_state.astype(self.observation_space.dtype)

    def reward(self, state: np.ndarray, act: int, next_state: np.ndarray) -> float:
        """Returns +1.0 reward when action is the right label and 0.0 otherwise."""
        n = self._length
        label = np.sum(state[: n // 2]) > np.sum(state[n // 2 :])
        return float(act == label)

    def transition(self, state: np.ndarray, action: int) -> np.ndarray:
        """Returns same state."""
        return state



"""Environment to sort a list using swap actions."""

from gym import spaces
import numpy as np

from seals import base_envs


class SortEnv(base_envs.ResettableMDP):
    """Environment to sort a list using swap actions."""

    def __init__(self, length: int = 4):
        """Constructs environment.

        The initial state is a vector x sampled uniformly from
        [0,1]**L, with actions a = (i,j) swapping x_i and x_j.  The
        reward is given according to the number of elements in the
        correct position.  To perform well, the learned policy must
        compare elements, otherwise it will not generalize to all
        possible randomly selected initial states.
        """
        self._length = length

        super().__init__(
            state_space=spaces.Box(low=0, high=1.0, shape=(length,)),
            action_space=spaces.MultiDiscrete([length, length]),
        )

    def terminal(self, state: np.ndarray, n_actions_taken: int) -> bool:
        """Always returns False."""
        return False

    def initial_state(self):
        """Sample random vector uniformly in [0, 1]**L."""
        sample = self.rand_state.random(size=self._length)
        return sample.astype(self.state_space.dtype)

    def reward(
        self,
        state: np.ndarray,
        action: np.ndarray,
        new_state: np.ndarray,
    ) -> float:
        """Rewards fully sorted lists, and new correct positions."""
        # This is not meant to be a potential shaping in the formal sense,
        # as it changes the trajectory returns (since we do not return
        # a fixed-potential state at termination).
        num_correct = self._num_correct_positions(state)
        new_num_correct = self._num_correct_positions(new_state)
        potential_diff = new_num_correct - num_correct

        return float(self._is_sorted(new_state)) + potential_diff

    def transition(self, state: np.ndarray, action: np.ndarray) -> np.ndarray:
        """Action a = (i, j) swaps elements in positions i and j."""
        new_state = state.copy()
        i, j = action
        new_state[[i, j]] = new_state[[j, i]]
        return new_state

    def _is_sorted(self, arr: np.ndarray) -> bool:
        return list(arr) == sorted(arr)

    def _num_correct_positions(self, arr: np.ndarray) -> int:
        return np.sum(arr == sorted(arr))



"""Benchmark environments for reward modeling and imitation."""

import gym

from seals import util
import seals.diagnostics  # noqa: F401
from seals.version import VERSION as __version__  # noqa: F401

# Classic control

gym.register(
    id="seals/CartPole-v0",
    entry_point="seals.classic_control:FixedHorizonCartPole",
    max_episode_steps=500,
)

gym.register(
    id="seals/MountainCar-v0",
    entry_point="seals.classic_control:mountain_car",
    max_episode_steps=200,
)

# MuJoCo

for env_base in ["Ant", "HalfCheetah", "Hopper", "Humanoid", "Swimmer", "Walker2d"]:
    gym.register(
        id=f"seals/{env_base}-v0",
        entry_point=f"seals.mujoco:{env_base}Env",
        max_episode_steps=util.get_gym_max_episode_steps(f"{env_base}-v3"),
    )



"""Environment testing for correct behavior under stochasticity."""

import numpy as np

from seals import base_envs


class RiskyPathEnv(base_envs.TabularModelMDP):
    """Environment with two paths to a goal: one safe and one risky.

    Many LfH algorithms are derived from Maximum Entropy Inverse Reinforcement
    Learning [1], which models the demonstrator as producing trajectories with
    probability p(tau) proportional to exp(R(tau)).  This model implies that a
    demonstrator can "control" the environment well enough to follow any
    high-reward trajectory with high probability [2]. However, in stochastic
    environments, the agent cannot control the probability of each trajectory
    independently.  This misspecification may lead to poor behavior.

    This task tests for this behavior. The agent starts at s_0 and can reach
    the goal s_2 (reward 1.0) by either taking the safe path s_0 to s_1 to s_2,
    or taking a risky action, which has equal chances of going to either s_3
    (reward -100.0) or s_2.  The safe path has the highest expected return, but
    the risky action sometimes reaches the goal s_2 in fewer timesteps, leading
    to higher best-case return. Algorithms that fail to correctly handle
    stochastic dynamics may therefore wrongly believe the reward favors taking
    the risky path.

    [1] Ziebart, Brian D., et al. "Maximum entropy inverse reinforcement
        learning." AAAI. Vol. 8. 2008.
    [2] Ziebart, Brian D. "Modeling purposeful adaptive behavior with the
        principle of maximum causal entropy." (2010); PhD thesis,
        CMU-ML-10-110; page 105.
    """

    def __init__(self):
        """Initialize environment."""
        nS = 4
        nA = 2

        transition_matrix = np.zeros((nS, nA, nS))
        transition_matrix[0, 0, 1] = 1.0
        transition_matrix[0, 1, [2, 3]] = 0.5

        transition_matrix[1, 0, 2] = 1.0
        transition_matrix[1, 1, 1] = 1.0

        transition_matrix[[2, 3], :, [2, 3]] = 1.0

        reward_matrix = np.array([0.0, 0.0, 1.0, -100.0])

        super().__init__(
            transition_matrix=transition_matrix,
            reward_matrix=reward_matrix,
        )



"""setup.py for seals project."""

import os
import sys

from setuptools import find_packages, setup  # type:ignore


def get_version() -> str:
    """Load version from version.py.

    Changes system path internally to avoid missing dependencies breaking imports.
    """
    sys.path.insert(
        0,
        os.path.join(os.path.dirname(__file__), "src", "seals"),
    )
    from version import VERSION  # type:ignore

    del sys.path[0]
    return VERSION


def get_readme() -> str:
    """Retrieve content from README."""
    with open("README.md", "r") as f:
        return f.read()


TESTS_REQUIRE = [
    # remove pin once https://github.com/nedbat/coveragepy/issues/881 fixed
    "black",
    "coverage==4.5.4",
    "codecov",
    "codespell",
    "darglint>=1.5.6",
    "flake8",
    "flake8-blind-except",
    "flake8-builtins",
    "flake8-commas",
    "flake8-debugger",
    "flake8-docstrings",
    "flake8-isort",
    "isort",
    "mypy",
    "pydocstyle",
    "pytest",
    "pytest-cov",
    "pytest-xdist",
    "pytype",
    "stable-baselines3>=0.9.0",
    "pyglet>=1.4.0",
]
DOCS_REQUIRE = [
    "sphinx",
    "sphinx-autodoc-typehints",
    "sphinx-rtd-theme",
    "sphinxcontrib-napoleon",
]

setup(
    name="seals",
    version=get_version(),
    description="Suite of Environments for Algorithms that Learn Specifications",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    author="Center for Human-Compatible AI",
    python_requires=">=3.7.0",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"seals": ["py.typed"]},
    install_requires=["gym"],
    tests_require=TESTS_REQUIRE,
    extras_require={
        # recommended packages for development
        "dev": ["ipdb", "jupyter", *TESTS_REQUIRE, *DOCS_REQUIRE],
        "docs": DOCS_REQUIRE,
        "test": TESTS_REQUIRE,
        # We'd like to specify `gym[mujoco]`, but this is a no-op when Gym is already
        # installed. See https://github.com/pypa/pip/issues/4957 for issue.
        "mujoco": ["mujoco_py>=1.50, <2.0", "imageio"],
    },
    url="https://github.com/HumanCompatibleAI/benchmark-environments",
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Environment testing for robustness to noise."""

from gym import spaces
import numpy as np

from seals import base_envs, util


class NoisyObsEnv(base_envs.ResettablePOMDP):
    """Simple gridworld with noisy observations.

    The agent randomly starts at the one of the corners of an MxM grid and
    tries to reach and stay at the center. The observation consists of the
    agent's (x,y) coordinates and L "distractor" samples of Gaussian noise .
    The challenge is to select the relevant features in the observations, and
    not overfit to noise.
    """

    def __init__(self, *, size: int = 5, noise_length: int = 20):
        """Build environment.

        Args:
            size: width and height of gridworld.
            noise_length: dimension of noise vector in observation.
        """
        self._size = size
        self._noise_length = noise_length
        self._goal = np.array([self._size // 2, self._size // 2])

        super().__init__(
            state_space=spaces.MultiDiscrete([size, size]),
            action_space=spaces.Discrete(5),
            observation_space=spaces.Box(
                low=np.concatenate(([0, 0], np.full(self._noise_length, -np.inf))),
                high=np.concatenate(
                    ([size - 1, size - 1], np.full(self._noise_length, np.inf)),
                ),
                dtype=np.float32,
            ),
        )

    def terminal(self, state: np.ndarray, n_actions_taken: int) -> bool:
        """Always returns False."""
        return False

    def initial_state(self) -> np.ndarray:
        """Returns one of the grid's corners."""
        n = self._size
        corners = np.array([[0, 0], [n - 1, 0], [0, n - 1], [n - 1, n - 1]])
        return corners[self.rand_state.randint(4)]

    def reward(self, state: np.ndarray, action: int, new_state: np.ndarray) -> float:
        """Returns  +1.0 reward if state is the goal and 0.0 otherwise."""
        return float(np.all(state == self._goal))

    def transition(self, state: np.ndarray, action: int) -> np.ndarray:
        """Returns next state according to grid."""
        return util.grid_transition_fn(
            state,
            action,
            x_bounds=(0, self._size - 1),
            y_bounds=(0, self._size - 1),
        )

    def obs_from_state(self, state: np.ndarray) -> np.ndarray:
        """Returns (x, y) concatenated with Gaussian noise."""
        noise_vector = self.rand_state.randn(self._noise_length)
        return np.concatenate([state, noise_vector]).astype(np.float32)
